/* denoising.h */
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright (C) 2007 Julien Mairal 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.
   */

/*!
 * \file
 *                Multiscale KSVD Denoising
 *
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                      File denoising.h
 * \brief Contains the class Denoising */

#ifndef DENOISING_H
#define DENOISING_H

#include "algo.h"
#include "processor.h"
#include "timer.h"

#define FACTOR_LAMBDA 0.46875
/// future functionality which prevents the usage of too much memory
#define MAX_ALLOCATE_RESIDUAL 4.5e9
#define ITER_WEIGHTED_LEARNING 10
#define ITER_LEARNING 1
#define DEFAULT_MAX_SPARSITY_DENOISING 120
#define NUM_ITER_VIDEO_INITIAL 30
#define MAX_SPARSITY_INITIAL_VIDEO 50
//#define PRUNE_RATIO_DENOISING 0.25
//#define SIZE_BLOCK_DENOISING 300
#define FACTOR_ITER_PROPAG_DICT 1

static double PRUNE_RATIO_DENOISING;
static int SIZE_BLOCK_DENOISING;
static int NN, choiceC;
using namespace std;

namespace KSVD {

   /*!
    * \brief This class implements the KSVD algorithm for denoising with
    * multiscale. 
    *
    * It is derivated from KSVD::Algo. T is the type of 
    * input data, V is the number of channels.
    **/
   template <typename T, int V> class Denoising : 
      public Algo<T,V> {
         public:
            /// Constructor, input is the noisy image, sigma is the
            /// standrad deviation of the noise, J is the number of
            /// iterations, fileDictionary is the file name of the 
            /// dictionary, original, the non-noisy image is optional,
            /// if it is provided, if will provide the PSNR during
            /// the process
            Denoising(const Image<T,V>* const input, double sigma,  
                  int J, 
                  const QString& fileDictionary, 
                  const Image<T,V>* const original = NULL,
                  int maxL = DEFAULT_MAX_SPARSITY_DENOISING,
                  int overlapFactor = OVERLAP_FACTOR);
            Denoising(const Image<T,V>* const input, double sigma,  
                  int J, 
                  Dictionary<T,V>* dictionary, 
                  const Image<T,V>* const original = NULL,
                  int maxL = DEFAULT_MAX_SPARSITY_DENOISING,
                  int overlapFactor = OVERLAP_FACTOR);
            /// Virtual Destructor
            virtual ~Denoising() { };
            /// print some information about the algorithm's parameters
            virtual void printInformation() const;
            /// execute the denoising algorithm
            virtual void exec(const T gam = GAMMA, const bool weightConstant = true);
            /// execute the fast denoising algorithm
            virtual void fastExec(const T gam = GAMMA);
            /// execute the Block denoising algorithm
            virtual void blockExec(const T gam = GAMMA, bool fastExec = true);

            /// Modifier
            void inline setFixedSparsity(bool newSpars) {
               _fixedSparsity = newSpars;
            };
            /// Modifier
            void inline setChoiceC(int choiceC) { _choiceC = choiceC;};

         protected:
            /// set the parameters _choiceC and the parameter N
            void setOptimalParameterN();
            /// Lagrange multiplier
            double _lambda;
            /// number of iterations
            int _J;
            /// deviation of the white gaussian noise
            double _sigma;
            /// permits to retrive the parameter C
            int _choiceC;
            /// max sparsity
            int _maxL;
            /// fixed sparsity
            bool _fixedSparsity;
      };

   /// Denoising algorithm for grayscale images
   typedef Denoising<double,1> DenoisingGray;
   /// Denoising algorithm for RGB images
   typedef Denoising<double,3> DenoisingRGB;

   /// Constructor, input is the noisy image, sigma is the
   /// standrad deviation of the noise, J is the number of
   /// iterations, fileDictionary is the file name of the 
   /// dictionary, original, the non-noisy image is optional,
   /// if it is provided, if will provide the PSNR during
   /// the process
   /// TODO factorize the code
   template <typename T, int V>
      Denoising<T,V>::Denoising(const Image<T,V>* const input, 
            double sigma, int J, 
            const QString& fileDictionary, 
            const Image<T,V>* const original,
            const int maxL, 
            const int overlapFactor) :
         Algo<T,V>(input,fileDictionary,original,overlapFactor),
         _sigma(sigma),
         _maxL(maxL),
         _fixedSparsity(false),
         _J(J) { 
            _lambda = ((this->_dictionary->sizeEdgeBigPatches()*
                     this->_dictionary->sizeEdgeBigPatches()/
                     (this->_overlapFactor*this->_overlapFactor))*
                  FACTOR_LAMBDA/sigma);
            setOptimalParameterN();
            /// if the image is too small, there will not be enough
            /// material for the learning process.
            if (this->_input->width() < 150 ||
                  this->_input->height() < 150) {
               cerr << "Warning : the input image is too small" << endl;
               cerr << "Risk of overfitting" << endl;
         /*      this->_N = this->_dictionary->N();
               this->_choiceC = 93;*/
               //               exit(1);
            }
         };

   template <typename T, int V>
      Denoising<T,V>::Denoising(const Image<T,V>* const input, 
            double sigma, int J, 
            Dictionary<T,V>* dictionary, 
            const Image<T,V>* const original,
            const int maxL,
            const int overlapFactor) :
         Algo<T,V>(input,dictionary,original,overlapFactor),
         _sigma(sigma),
         _maxL(maxL),
         _fixedSparsity(false),
         _J(J) { 
            _lambda = ((this->_dictionary->sizeEdgeBigPatches()*
                     this->_dictionary->sizeEdgeBigPatches()/
                     (this->_overlapFactor*this->_overlapFactor))*
                  FACTOR_LAMBDA/sigma);
            setOptimalParameterN();
            /// if the image is too small, there will not be enough
            /// material for the learning process.
            if (this->_input->width() < 150 ||
                  this->_input->height() < 150) {
               cerr << "Warning : the input image is too small" << endl;
               cerr << "Risk of overfitting" << endl;
      //         this->_N = this->_dictionary->N();
         //      this->_choiceC = 93;
               //               exit(1);
            }
         };


   /// print some information about the algorithm's parameters.
   template <typename T, int V>
      void Denoising<T,V>::printInformation() const {
         cout << "Executing Denoising Algorithm with Multi-Resolution"
            << " KSVD" << endl;
         cout << "Sigma = " << _sigma << endl;
         cout << "J = " << _J << endl;
         Algo<T, V>::printInformation();
      };

   /// execute the denoising algorithm.
   template <typename T, int V> 
      void Denoising<T,V>::exec(const T gam, const bool weightConstant) {
         list<int>* listPatches = new list<int>();
#ifdef _OPENMP
         /// adapt the number of threads to the number of processors
         /// available
         int NUM_THREADS =  MIN(4,omp_get_num_procs());
#else 
         int NUM_THREADS = 1;
#endif
         int numPatches = this->generateListPatches(*listPatches);
         /// if RGB space is selected, use the RGB metric for denoising
        // T gamma = V==3 ? gam: 0;
         T gamma = gam;
         if (gamma > EPSILON) this->_dictionary->switchToRGBMetric(gamma);
         cout << "Gamma : " << gamma << endl;
         /// this boolean is related to a future functionality 
         /// which will reduce drastically the memory occupation
         bool storeResidual = _J != 0 && (numPatches*
               this->_dictionary->n(this->_dictionary->N()-1)) <
            MAX_ALLOCATE_RESIDUAL;
         /// storeResidual functionality has not been implemented yet
         /// so it is set to true by default
         storeResidual = true;
         /// allocate the workspace memory for the algorithm
         this->_dictionary->allocateLearning(numPatches,this->_N,
               _fixedSparsity,
               storeResidual);
         cout << "Iteration n0" << endl;
         /// one class Processor for each thread
         Processor<T,V>** processors = new Processor<T,V>*[NUM_THREADS];
         for (int i = 0; i<NUM_THREADS; ++i) {
            processors[i] = new Processor<T,V>(
                  this->_dictionary,_sigma, this->_N,_choiceC,gamma,
                  this->_input,_maxL);
         }
         Timer clock;
         clock.start();
         this->approximation(processors,*listPatches,NUM_THREADS, V <= 9 && weightConstant);
         if (this->_stats)
            this->_dictionary->printSparsity();
         /// Principal loop
         for (int i = 1; i<=_J; ++i) {
            clock.pause();
            /// Do not take into account the time spent for 
            /// statistical analysis
            if (this->_stats && this->_original != NULL) {
               this->generateOutput(*listPatches, _lambda);
               this->printPSNR();
            }
            clock.restart();

            /// go back to the algorithm
            cout << "Iteration n" << i << endl;
            for (int j = 0; j<ITER_LEARNING; ++j) {
               this->_dictionary->performLearning(2*i <= _J &&
                     j == ITER_LEARNING-1, 2*i==_J || 2*i ==_J+1,
                     this->_input);
            }
            if (gamma > EPSILON) 
               this->_dictionary->switchToRGBMetric(gamma);
            approximation(processors,*listPatches,NUM_THREADS,
                  V <= 9 && weightConstant);
            if (this->_stats)
               this->_dictionary->printSparsity();
         }

         this->generateOutput(*listPatches, _lambda);
         this->printPSNR();
         /// deallocate the workspace for the learning
         this->_dictionary->terminateLearning();
         for (int i = 0; i<NUM_THREADS; ++i) {
            delete(processors[i]);
            processors[i] = NULL;
         }
         delete[](processors);

         clock.stop();
         clock.printTimeElapsed();
         delete(listPatches);
         listPatches = NULL;
      };

   /// execute the fast denoising algorithm.
   template <typename T, int V> 
      void Denoising<T,V>::fastExec(const T gam) {
         list<int>* listPatches = new list<int>();
#ifdef _OPENMP
         /// adapt the number of threads to the number of processors
         /// available
         int NUM_THREADS =  omp_get_num_procs();
#else 
         int NUM_THREADS = 1;
#endif
         int numPatches = this->generateListPatches(*listPatches);
         /// if RGB space is selected, use the RGB metric for denoising
         T gamma = gam;
         if (gamma > EPSILON) this->_dictionary->switchToRGBMetric(gamma);
         cout << "Gamma : " << gamma << endl;
         /// this boolean is related to a future functionality 
         /// which will reduce drastically the memory occupation
         bool storeResidual = _J != 0 && (numPatches*
               this->_dictionary->n(this->_dictionary->N()-1)) <
            MAX_ALLOCATE_RESIDUAL;
         /// storeResidual functionality has not been implemented yet
         /// so it is set to true by default
         storeResidual = true;
         /// one class Processor for each thread
         Processor<T,V>** processors = new Processor<T,V>*[NUM_THREADS];
         for (int i = 0; i<NUM_THREADS; ++i) {
            processors[i] = new Processor<T,V>(
                  this->_dictionary,_sigma, this->_N,_choiceC,gamma,
                  this->_input,_maxL);
         }
 
         /// Learning step
         cout << "Learning on a random subset of patches" << endl;
         Timer clock;
         clock.start();
         /// Principal loop
         for (int i = 1; i<=_J; ++i) {
            /// go back to the algorithm
            cout << "Iteration n" << i << endl;
            list<int>* prunedListPatches = new list<int>(*listPatches);
            pruneListPatches(*prunedListPatches, PRUNE_RATIO_DENOISING);
            int prunedNumPatches = prunedListPatches->size();

            /// allocate the workspace memory for the algorithm
            this->_dictionary->allocateLearning(prunedNumPatches,this->_N,
                  _fixedSparsity, storeResidual);
            this->approximation(processors,*prunedListPatches,NUM_THREADS, V <= 3);

            for (int j = 0; j<ITER_LEARNING; ++j) {
               this->_dictionary->performLearning(2*i <= _J &&
                     j == ITER_LEARNING-1, 2*i==_J || 2*i ==_J+1,
                     this->_input);
            }
            if (gamma > EPSILON) 
               this->_dictionary->switchToRGBMetric(gamma);

            /// deallocate the workspace for the learning
            this->_dictionary->terminateLearning();
            delete(prunedListPatches);
         }

         cout << "Reconstruction" << endl;
         this->_dictionary->allocateLearning(numPatches,this->_N,
               _fixedSparsity, storeResidual);
         this->approximation(processors,*listPatches,NUM_THREADS, V <= 3);
         this->_dictionary->printSparsity();
         this->generateOutput(*listPatches, _lambda);
         this->printPSNR();

         /// deallocate the workspace for the learning
         this->_dictionary->terminateLearning();
         for (int i = 0; i<NUM_THREADS; ++i) {
            delete(processors[i]);
            processors[i] = NULL;
         }
         delete[](processors);

         clock.stop();
         clock.printTimeElapsed();
         delete(listPatches);
         listPatches = NULL;
      };

   /// execute the block fast denoising algorithm.
   template <typename T, int V> 
      void Denoising<T,V>::blockExec(const T gam, const bool fast) {
         Timer clock;
         clock.start();
         int tmpX = this->_input->height()-SIZE_BLOCK_DENOISING;
         int tmpY = this->_input->width()-SIZE_BLOCK_DENOISING;
         int sizeBlock = SIZE_BLOCK_DENOISING-
            this->_dictionary->sizeEdgeBigPatches();

         int numBlockX,numBlockY,sizeBlockDenoisingX,sizeBlockDenoisingY;
         if (tmpX < 0) {
            numBlockX = 1;
            sizeBlockDenoisingX = this->_input->height();
         } else {
            numBlockX = tmpX/sizeBlock+1;
            if (!(tmpX % sizeBlock == 0)) ++numBlockX;
            sizeBlockDenoisingX = SIZE_BLOCK_DENOISING;
         }
         if (tmpY < 0) {
            numBlockY = 1;
            sizeBlockDenoisingY = this->_input->width();
         } else {
            numBlockY = tmpY/sizeBlock+1;
            if (!(tmpY % sizeBlock == 0)) ++numBlockY;
            sizeBlockDenoisingY = SIZE_BLOCK_DENOISING;
         }

         delete(this->_output);
         this->_output = new Image<T,V>(this->_input->width(),
               this->_input->height());
         *this->_output = 0.0;
         Image<T,V>* weights = new Image<T,V>(*this->_output);
         Image<T,V>* blockOnes = new Image<T,V>(sizeBlockDenoisingY,
               sizeBlockDenoisingX);
         *blockOnes = 1.0;
         int count = 0;

         for (int i = 0; i<numBlockX; ++i) {
            for (int j = 0; j<numBlockY; ++j) {
               cout << "Block " << count << " is being processed" << endl;
               int cornerX = MAX(0,MIN(i*sizeBlock,
                     this->_input->height()-sizeBlockDenoisingX));
               int cornerY = MAX(0,MIN(j*sizeBlock,
                     this->_input->width()-sizeBlockDenoisingY));
               Image<T,V>* blockInput = this->_input->regionOfInterest(
                     cornerX,cornerY,
                     sizeBlockDenoisingX,sizeBlockDenoisingY);
         /*      Image<T,V>* blockOriginal = this->_original->regionOfInterest(
                     cornerX,cornerY,
                     sizeBlockDenoisingX,sizeBlockDenoisingY);*/
               Denoising<T,V>* algo = new Denoising<T,V>(blockInput,_sigma,
                     count == 0 ? _J : _J/FACTOR_ITER_PROPAG_DICT
                     ,this->_dictionary,NULL,_maxL,this->_overlapFactor);
               if (fast) {
                  algo->fastExec(gam);
               } else {
                  algo->exec(gam);
               }
               this->_output->addToRegionOfInterest(algo->output(),
                     cornerX,cornerY);
               weights->addToRegionOfInterest(*blockOnes,cornerX,cornerY);
               algo->setDictionaryNull();
               delete(algo);
               delete(blockInput);
/*               delete(blockOriginal);*/
               ++count;
            }
         }
         *(this->_output) /= *weights;
         delete(weights);
         delete(blockOnes);
         clock.stop();
         clock.printTimeElapsed();
      };

   /// set the optimal parameter N
   template <typename T, int V>
      void Denoising<T,V>::setOptimalParameterN() {
         /// These parameters have been tuned for N = 1,2, V = 1
         if (V == 1) {
            switch (this->_N) {
               case 1: if (_sigma < 7) {
                          this->_N = 1;
                          _choiceC = 93;
                       } else if (_sigma < 12) {
                          this->_N = 1;
                          _choiceC = 93;
                       } else if (_sigma < 17) {
                          this->_N = 4;
                          _choiceC = 999;
                       } else if (_sigma < 35) {
                          this->_N = 4;
                          _choiceC = 99;
                       } else {
                          this->_N = 4;
                          _choiceC = 95;
                       }
                       break;
               case 2: if (_sigma < 7) {
                          this->_N = 3;
                          _choiceC = 999;
                       } else if (_sigma < 12) {
                          this->_N = 3;
                          _choiceC = 97;
                       } else if (_sigma < 17) {
                          this->_N = 4;
                          _choiceC = 99;
                       } else if (_sigma < 23) {
                          this->_N = 4;
                          _choiceC = 99;
                       } else if (_sigma < 27) {
                          this->_N = 4;
                          _choiceC = 97;
                       } else if (_sigma < 60) {
                          this->_N = 5;
                          _choiceC = 90;
                       } else {
                          this->_N = 5;
                          _choiceC = 85;
                       }
                       break;
               default:

                       this->_N = NN;
                       _choiceC = choiceC;
                       /* this->_N = this->_dictionary->N(); 
                          _choiceC = 93;*/ break;
            }
         } else if (V == 3) {
            // this->_N = this->_dictionary->N();
            // _choiceC = 93;
            this->_N = NN;
            _choiceC = choiceC;
         } else {
            this->_N = NN;
            _choiceC = choiceC;
          //  this->_N = 4;
          //  _choiceC = 97;
            //   this->_N = this->_dictionary->N();
            //   _choiceC = 93;
         }
      };

   /*! 
    * namespace for weighted K-SVD algorithms
    * */
   namespace Weighted {

      template <typename T, int V> class Denoising : 
         public KSVD::Denoising<T,V> {
            public:
               /// Constructor
               Denoising(const Image<T,V>* const input, 
                     const Image<T,V>& mapSigma,  
                     int J, const QString& fileDictionary, 
                     const Image<T,V>* const original = NULL);
               /// Destructor
               virtual ~Denoising() { delete(_mapSigma); }

               /// print some information about the algorithm's parameters
               virtual void printInformation() const;
               /// execute the denoising algorithm
               void exec(const T gam = GAMMA, const bool weightConstant = true);

            private:
               /// matrix \f$ \beta \f$:w
               Image<T,V>* _mapSigma;
         };

      /// Denoising algorithm for grayscale images
      typedef Denoising<double,1> DenoisingGray;
      /// Denoising algorithm for RGB images
      typedef Denoising<double,3> DenoisingRGB;

      /// print some information about the algorithm's parameters.
      template <typename T, int V>
         void Denoising<T,V>::printInformation() const {
            cout << "Executing Weighted Denoising Algorithm with "
               << "Multiscale KSVD" << endl;
            cout << "Sigma Minimum = " << this->_sigma << endl;
            cout << "J = " << this->_J << endl;
            Algo<T,V>::printInformation();
         };

      /// Constructor
      template <typename T, int V>
         Denoising<T,V>::Denoising(const Image<T,V>* const input, 
               const Image<T,V>& mapSigma, int J, 
               const QString& fileDictionary, 
               const Image<T,V>* const original) :
            KSVD::Denoising<T,V>::Denoising(input,1,J,fileDictionary,
                  original) {
               _mapSigma = new Image<T,V>(mapSigma);
               this->_sigma = MAX(EPSILON_IMAGE,_mapSigma->min());
               _mapSigma->inv();
               (*_mapSigma) *= this->_sigma;
               this->_lambda = ((this->_dictionary->sizeEdgeBigPatches()*
                        this->_dictionary->sizeEdgeBigPatches()/
                        (this->_overlapFactor*this->_overlapFactor))*
                     FACTOR_LAMBDA/this->_sigma);
               this->_N = 1; 
               this->_choiceC = 93; 
            };

      /// execute the denoising algorithm.
      template <typename T, int V> 
         void Denoising<T,V>::exec(const T gam, const bool weightConstant) {
            list<int>* listPatches = new list<int>();
#ifdef _OPENMP
            /// adapt the number of threads to the number of processors
            /// available
            int NUM_THREADS =  omp_get_num_procs();
#else 
            int NUM_THREADS = 1;
#endif
            int numPatches = this->generateListPatches(*listPatches);
            /// if RGB space is selected, use the RGB metric for denoising
            T gamma = V == 3 ? gam : 0;
            if (gamma > EPSILON) 
               this->_dictionary->switchToRGBMetric(gamma);
            /// this boolean is related to a future functionality 
            /// which will reduce drastically the memory occupation
            bool storeResidual = this->_J != 0 && (numPatches*
                  this->_dictionary->n(this->_dictionary->N()-1)) <
               MAX_ALLOCATE_RESIDUAL;
            /// storeResidual functionality has not been implemented yet
            /// so it is set to true by default
            storeResidual = true;
            /// allocate the workspace memory for the algorithm
            this->_dictionary->allocateLearning(numPatches,this->_N,
                  this->_fixedSparsity,
                  storeResidual);
            cout << "Iteration n0" << endl;
            /// one class Processor for each thread
            KSVD::Processor<T,V>** processors = 
               new KSVD::Processor<T,V>*[NUM_THREADS];
            for (int i = 0; i<NUM_THREADS; ++i) {
               processors[i] = new Processor<T,V>(
                     this->_dictionary, this->_sigma, this->_N,
                     this->_choiceC,gamma,
                     this->_input,this->_mapSigma,this->_maxL);
            }
            Timer clock;
            clock.start();
            this->approximation(processors,*listPatches,NUM_THREADS,
                  V<=3 && weightConstant);
            this->_dictionary->printSparsity();
            /// Principal loop
            for (int i = 1; i<=this->_J; ++i) {
               clock.pause();
               /// Do not take into account the time spent for 
               /// statistical analysis
               this->generateOutput(*listPatches, this->_lambda,_mapSigma);
               this->printPSNR();
               clock.restart();

               /// go back to the algorithm
               cout << "Iteration n" << i << endl;
               for (int j = 0; j<ITER_WEIGHTED_LEARNING; ++j)
                  this->_dictionary->performLearning(2*i <= this->_J &&
                        j == ITER_WEIGHTED_LEARNING-1, 2*i == this->_J || 2*i ==this->_J+1 ,
                        this->_input,_mapSigma);
               if (gamma > EPSILON) 
                  this->_dictionary->switchToRGBMetric(gamma);
               this->approximation(processors,*listPatches,NUM_THREADS,
                     V <=3 && weightConstant);
               this->_dictionary->printSparsity();
            }

            this->generateOutput(*listPatches, this->_lambda,_mapSigma);
            this->printPSNR();
            /// deallocate the workspace for the learning
            this->_dictionary->terminateLearning();
            for (int i = 0; i<NUM_THREADS; ++i) {
               delete(processors[i]);
               processors[i] = NULL;
            }
            delete[](processors);

            clock.stop();
            clock.printTimeElapsed();
            delete(listPatches);
            listPatches = NULL;

         };
   };

   namespace Sequence {
      /*!
       * \brief This class implements the KSVD algorithm for denoising Videos
       * with multiscale. 
       *
       * It is derivated from KSVD::Algo. T is the type of 
       * input data, V is the number of channels, W is the number of frames
       * in the 3D patches
       **/
      template <typename T = double, int V = 1, int W = 3> class Denoising : 
         public Algo<T,V,W> {
            public:
               /// Constructor, input is the noisy image, sigma is the
               /// standrad deviation of the noise, J is the number of
               /// iterations, fileDictionary is the file name of the 
               /// dictionary, original, the non-noisy image is optional,
               /// if it is provided, if will provide the PSNR during
               /// the process
               Denoising(const Video<T,V,W>* const input, double sigma,  
                     int J, 
                     const QString& fileDictionary, 
                     const Video<T,V,W>* const original = NULL,
                     int maxL = DEFAULT_MAX_SPARSITY_DENOISING,
                     const int overlapFactor = OVERLAP_FACTOR_VIDEO);
               /// Virtual Destructor
               virtual ~Denoising() { };
               /// print some information about the algorithm's parameters
               virtual void printInformation() const;
               /// execute the denoising algorithm
               void exec(const T gam, const bool fast,
                     const bool block = false);
               void exec(const T gam = GAMMA) {
                  exec(gam,false,false);
               };

            protected:
               /// number of iterations
               int _J;
               /// deviation of the white gaussian noise
               double _sigma;
               /// max sparsity
               int _maxL;
         };

      /// Constructor, input is the noisy image, sigma is the
      /// standrad deviation of the noise, J is the number of
      /// iterations, fileDictionary is the file name of the 
      /// dictionary, original, the non-noisy image is optional,
      /// if it is provided, if will provide the PSNR during
      /// the process
      template <typename T, int V, int W>
         Denoising<T,V,W>::Denoising(const Video<T,V,W>* const input, 
               double sigma,  
               int J, 
               const QString& fileDictionary, 
               const Video<T,V,W>* const original,
               int maxL,
               const int overlapFactor) : 
            Algo<T,V,W>(input,fileDictionary,original),
            _sigma(sigma),
            _maxL(maxL), _J(J) {
               /// if the image is too small, there will not be enough
               /// material for the learning process.
               if (this->_input->width() < 150 ||
                     this->_input->height() < 150) {
                  cerr << "Warning : the input image is too small" << endl;
                  cerr << "Risk of overfitting" << endl;
                  //               exit(1);
               }
            };

      /// execute the denoising algorithm
      template <typename T, int V, int W>
         void Denoising<T,V,W>::exec(const T gam, bool fast, bool block) {
            this->_output = new Video<T,V,W>(this->_input->width(),
                  this->_input->height(), this->_input->nFrames());
            this->_output->clear();
            const int sizeFrame = this->_input->width()*
               this->_input->height()*V;
            const int offset = W/2;
            KSVD::Denoising<T,V*W>* algorithm;
            for (int i = offset;
                  i< this->_input->nFrames()-offset; ++i) {
               cout << "********************************************"
                  << endl;
               cout << "Frame #" << i << " is being processed" << endl;
               const_cast<Video<T,V,W>*>(
                     this->_input)->setCurrentFrame(i);
               Image<T,V*W>* original = NULL;
               if (this->_original) {
                  const_cast<Video<T,V,W>*>(
                        this->_original)->setCurrentFrame(i);
                  original = this->_original->currentFrames();
               }
               algorithm = new KSVD::Denoising<T,V*W>(
                     this->_input->currentFrames(), _sigma, 
                     i == offset ? NUM_ITER_VIDEO_INITIAL : _J, 
                     this->_dictionary,original,i == offset ? 
                     MAX_SPARSITY_INITIAL_VIDEO : DEFAULT_MAX_SPARSITY_DENOISING,
                     this->_overlapFactor);
               if (block) {
                  algorithm->blockExec(gam,fast);
               } else {
                  if (fast) {
                     algorithm->fastExec(gam);
                  } else {
                     algorithm->exec(gam);
                  }
               }

               const Image<T,V*W>& output = algorithm->output();
               for (int j = 0; j<W; ++j) {
                  for (int l = 0; l<this->_input->width()*
                        this->_input->height(); ++l) {
                     for (int k = 0; k<V; ++k) {
                        (*this->_output)[(i-offset+j)*sizeFrame+l*V+k] +=
                           output[W*V*l+j*V+k];
                     }
                  }
               }
               algorithm->setDictionaryNull();
               delete(algorithm);
               algorithm = NULL;
            }
            cout << "********************************************"
               << endl;
            /// finish the averaging
            int boundary = MIN(W-1,this->_input->nFrames()-W);
            for (int i = 1; i<boundary; ++i) {
               for (int j = 0; j<sizeFrame; ++j) {
                  (*this->_output)[i*sizeFrame+j] /= (i+1);
               }
            }
            for (int i = boundary; i<this->_input->nFrames()-boundary; ++i) {
               for (int j = 0; j<sizeFrame; ++j) {
                  (*this->_output)[i*sizeFrame+j] /= boundary+1;
               }
            }
            for (int i = this->_input->nFrames()-boundary; 
                  i<this->_input->nFrames()-1; ++i) {
               for (int j = 0; j<sizeFrame; ++j) {
                  (*this->_output)[i*sizeFrame+j] /= 
                     (this->_input->nFrames()-i);
               }
            }
         };

      /// print some information about the algorithm's parameters.
      template <typename T, int V, int W>
         void Denoising<T,V,W>::printInformation() const {
            cout << "Executing Denoising Algorithm with Multi-Resolution"
               << " KSVD for Video" << endl;
            cout << "Sigma = " << _sigma << endl;
            cout << "J = " << _J << endl;
            Algo<T,V,W>::printInformation();
         };
   };

   typedef KSVD::Sequence::Denoising<double,1,3> DenoisingVideoGray;
   typedef KSVD::Sequence::Denoising<double,1,5> DenoisingHQVideoGray;
   typedef KSVD::Sequence::Denoising<double,3,3> DenoisingVideoRGB;
   typedef KSVD::Sequence::Denoising<double,3,5> DenoisingHQVideoRGB;
};

#endif // DENOISING_H
