/* processor.h */
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright (C) 2007 Julien Mairal 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.
*/

/*!
 * \file
 *                Multiscale KSVD Denoising
 *
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                      File processor.h
 * \brief Contains the class Processor */

#ifndef PROCESSOR_H
#define PROCESSOR_H

#include "dictionary.h"
#include "misc.h"
#include <qapplication.h>

//#define FILENAME_PARAMETER_C "misc/parameterC"
#define MAX_DIM 9000
#define DEFAULT_MAX_SPARSITY 120
#define QUOTEAUX(x) #x
#define QUOTE(x) QUOTEAUX(x)
//#define WEIGHT_CONSTANT 2.5
static double WEIGHT_CONSTANT;

namespace KSVD {

   /*! \brief This class makes a multiscale orthogonal matching pursuit.
    *
    * This is designed to be multi-thread */
   template <typename T, int V>
      class Processor {
         public:
            /// Constructor
            Processor(Dictionary<T,V>* dictionary, double sigma, int N,
                  const int choiceC,
                  const T gamma,
                  const Image<T,V>* const image = NULL,
                  const int maxL = DEFAULT_MAX_SPARSITY,
                  const bool weightConstant = true);
            /// Destructor
            virtual ~Processor();

            /// initialize the processor.
            virtual void initialize(int indexInCoeffs, int indexInImage);
            /// initialize the processor with a given patch
            virtual void initialize(int indexInCoeffs, 
                  const VectR& patch);
            /// launch the approximation
            void run();

            /// Modifier
            void inline setWeightConstant(const bool wc) {
               _weightConstant = wc && _dictionary->addConstantAtoms();
            };
            /// Accessor
            void inline weightConstant() const { return _weightConstant;};

         protected:
            /// Copy constructor is forbidden
            explicit Processor(const Processor<T,V>& process);
            /// operator= is forbidden
            Processor<T,V>& operator=(const Processor<T,V>& process);

            /// auxiliary function for initialization
            void initializeAux();
            /// approximation for a node
            void nodeApproximation(const int indexNode);
            /// approximation for the root
            void rootApproximation(const int level, const int node,
                  T* normsRoot);
            /// approximation for the root
            void rootApproximation();
            /// update the dictionary
            void updateDictionary();
            /// update the criterion for one root and its children
            inline void updateCriterion(const int indexNode, 
                  const int level);
            /// update the scores for each node and select the best
            /// level and atom
            inline void updateScores(const int indexNode, 
                  const int level, int& levelSelected,int& nodeSelected,
                  int& atomSelected);
            /// update the scores for each node and select the best
            /// level and atom
            inline void updateScores(const int indexNode, 
                  const int level);
            /// compute the coefficients alpha
            inline void computeCoeffs();
            /// update the matrices Q,R and the residual
            inline void computeQRResidual(const int indexNode,
                  const int levelSelected, const int nodeSelected,
                  const int atomSelected);
            /// update the matrices Q,R and the residual
            inline void computeQRResidual(const int indexNode);

            /// Dictionary
            Dictionary<T,V>* _dictionary;
            /// Image
            const Image<T,V>* const _image;

            /// give more weight to the constant atoms
            bool _weightConstant;
            /// gamma for RGB metric
            T _gamma;
            /// index of the patch processed
            int _indexPatch;
            /// index of the patch processed in the image
            int _indexInImage;
            /// nombre d'chelles actives
            const int _Na;
            /// nombre d'chelles totales
            int _Nt;
            /// maximum sparsity
            int _maxL;
            /// permits to retrieve the parameterC
            int _choiceC;
            /// sizes of the patches for each scale
            int* _n;
            /// sizes of the edges of the patches for each scale
            int* _sizeEdges;
            /// number of nodes
            int* _nNodes;
            /// number of roots
            int _nRoots;
            /// thrshold for each scale.
            T** _thrshold;
            /// converged or not for each root
            bool* _converged;
            /// score for each node.
            T** _score;
            /// Matrix Q for each root
            Matrix** _Q;
            /// Matrix R for each root
            Matrix** _R;
            /// number of selected atoms for each node
            int* _numSelectedAtoms;
            /// index of selected atoms for each root
            int** _indexSelectedAtoms;
            /// level of selected atoms for each root
            int** _levelOfSelectedAtoms;
            /// index of selected patches for each root
            int** _selectedPatches;
            /// selectionCriterion (scalarProducts if OMP)
            VectR*** _selectionCriterion;
            /// coefficients for each root
            T** _coeffs;
            /// inverse of the norm of a modified dictionary
            T*** _invNorm;
            /// residual for each node (share the same space memory)
            VectR*** _residuals;
            /// original patch for each root
            VectR** _originalPatch;
            /// dictionaries for each node
            Matrix*** _dicts;
      };

   /// Constructor
   template <typename T, int V>
      Processor<T,V>::Processor(Dictionary<T,V>* dictionary, 
            double sigma, int N, const int choiceC,
            const T gamma,
            const Image<T,V>* const image,
            const int maxL,
            const bool weightConstant) :
         _image(image),
         _gamma(gamma),
         _maxL(maxL),
         _choiceC(choiceC),
         _dictionary(dictionary),
         _Nt(N),
         _indexPatch(0),
         _Na(dictionary->N()) {
            _weightConstant = weightConstant && _dictionary->addConstantAtoms();
            if (_Na > _Nt) {
               cerr << "incoherent number of scales" << endl;
               exit(1);
            }
            _n = new int[_Nt];
            _n[0] = _dictionary->n(0);
            for (int i = 1; i<_Nt; ++i) _n[i] = 4*_n[i-1];
            _sizeEdges = new int[_Nt];
            for (int i = 0; i<_Nt; ++i) {
               _sizeEdges[i] = static_cast<int>(
                     sqrt(static_cast<double>(_n[i]/V)));
            }
            _nNodes = new int[_Nt];
            _nNodes[_Nt-1] = 1;
            for (int i = _Nt-2; i>=0; --i) {
               _nNodes[i] = 4*_nNodes[i+1];
            }
            _nRoots = _nNodes[_Na-1];
            _converged = new bool[_nRoots];
            _score = new T*[_Na];
            for (int i = 0; i<_Na; ++i) {
               _score[i] = new T[_nNodes[i]];
            }
            _numSelectedAtoms = new int[_nRoots];
            for (int j = 0; j<_nRoots; ++j) {
               _numSelectedAtoms[j] = 0;
            }
            _indexSelectedAtoms = new int*[_nRoots];
            for (int j = 0; j<_nRoots; ++j) {
               _indexSelectedAtoms[j] = 
                  new int[_maxL];
            }
            _selectedPatches = new int*[_nRoots];
            for (int j = 0; j<_nRoots; ++j) {
               _selectedPatches[j] = 
                  new int[_maxL];
            }
            _levelOfSelectedAtoms = new int*[_nRoots];
            for (int j = 0; j<_nRoots; ++j) {
               _levelOfSelectedAtoms[j] = 
                  new int[_maxL];
            }
            _coeffs = new T*[_nRoots];
            for (int i = 0; i<_nRoots; ++i) {
               _coeffs[i] = new T[_maxL];
            }
            _selectionCriterion = new VectR**[_Na];
            for (int i = 0; i<_Na; ++i) {
               _selectionCriterion[i] = new VectR*[_nNodes[i]];
               for (int j = 0; j<_nNodes[i]; ++j) {
                  _selectionCriterion[i][j] = 
                     new VectR(_dictionary->K(i));
               }
            }

            // set thrshold
            _thrshold = new T*[_Nt];
            for (int i = 0; i<_Nt; ++i) {
               _thrshold[i] = new T[_nNodes[i]];
            }
            std::ifstream f;
            // if choicC == -1, _C is fixed to 1
            if (_choiceC != -1) { 
               f.open(QString(QUOTE(FILENAME_PARAMETER_C))+QString::number(_choiceC)+
                     ".data");
               if (!f) {
                  cerr << "Problem while opening " << QUOTE(FILENAME_PARAMETER_C) << endl;
                  exit(1);
               }
            }
            double _C = 1.0;;
            int index = 0;
            for (int i = 1; i <= MIN(MAX_DIM,_n[_Nt-1]); ++i) {
               if (_choiceC != -1) f >> _C;
               if (_n[index] == i) {
                  _thrshold[index][0] = MAX(sqrt(static_cast<T>(
                              _n[index]))*_C*sigma, EPSILON);
                  for (int j = 1; j<_nNodes[index]; ++j) 
                     _thrshold[index][j] = _thrshold[index][0];
                  ++index;
               }
            }
            for (; index<_Nt; ++index) {
               _thrshold[index][0] = MAX(sqrt(static_cast<T>(
                           _n[index]))*_C*sigma,EPSILON);
               for (int j = 1; j<_nNodes[index]; ++j) 
                  _thrshold[index][j] = _thrshold[index][0];
            }
            if (_choiceC != -1) f.close();

            // set the structure for the residuals with shared memory
            _residuals = new VectR**[_Na];
            for (int i = 0; i<_Na; ++i) {
               _residuals[i] = new VectR*[_nNodes[i]];
            }
            for (int i = 0; i<_nRoots; ++i) {
               _residuals[_Na-1][i] = new VectR(_n[_Na-1]);
            }
            for (int i = _Na-2; i >= 0; --i) {
               for (int j = 0; j<_nNodes[i+1]; ++j) {
                  for (int k = 0; k<4; ++k) {
                     _residuals[i][4*j+k] = new VectR(_n[i],
                           *(_residuals[i+1][j]),k);
                  }
               }
            }      
            _originalPatch = new VectR*[_nRoots];
            for (int i = 0; i<_nRoots; ++i) {
               _originalPatch[i] = new VectR(_n[_Na-1]);
            }

            /// initialize the pointers for the matrix Q and R
            _Q = new Matrix*[_nRoots];
            for (int i = 0; i<_nRoots; ++i) _Q[i] = NULL;
            _R = new Matrix*[_nRoots];
            for (int i = 0; i<_nRoots; ++i) _R[i] = NULL;

            /// dictionaries for each node
            _dicts = new Matrix**[_Na];
            for (int i = 0; i<_Na; ++i) {
               _dicts[i] = new Matrix*[_nNodes[i]];
               for (int j = 0; j<_nNodes[i]; ++j) {
                  _dicts[i][j] = NULL;
               }
            }
            _invNorm = new T**[_Na];
            for (int i = 0; i<_Na; ++i) {
               _invNorm[i] = new T*[_nNodes[i]];
            }
         };

   /// Destructor
   template <typename T, int V>
      Processor<T,V>::~Processor() {
         delete[](_converged);
         for (int i = 0; i<_Na; ++i) {
            delete[](_score[i]);
            _score[i] = NULL;
         }
         for (int i = 0; i<_Na; ++i) {
            for (int j = 0; j<_nNodes[i]; ++j) 
               _invNorm[i][j] = NULL;
            delete[](_invNorm[i]);
            _invNorm[i] = NULL;
         }
         delete[](_invNorm);
         delete[](_score);
         for (int i = 0; i<_nRoots; ++i) {
            delete(_Q[i]);
            _Q[i] = NULL;
         }
         delete[](_Q);
         for (int i = 0; i<_nRoots; ++i) {
            delete(_R[i]);
            _R[i] = NULL;
         }
         delete[](_R);
         delete[](_numSelectedAtoms);
         for (int i = 0; i<_nRoots; ++i) {
            delete(_originalPatch[i]);
            _originalPatch[i] = NULL;
            delete[](_indexSelectedAtoms[i]);
            _indexSelectedAtoms[i] = NULL;
            delete[](_selectedPatches[i]);
            _selectedPatches[i] = NULL;
            delete[](_levelOfSelectedAtoms[i]);
            _levelOfSelectedAtoms[i] = NULL;
            delete[](_coeffs[i]);
            _coeffs[i] = NULL;
         }
         delete[](_indexSelectedAtoms);
         delete[](_selectedPatches);
         delete[](_levelOfSelectedAtoms);
         delete[](_coeffs);
         delete[](_originalPatch);
         for (int i = 0; i<_Na; ++i) {
            for (int j = 0; j<_nNodes[i]; ++j) {
               delete(_selectionCriterion[i][j]);
               _selectionCriterion[i][j] = NULL;
            }
            delete[](_selectionCriterion[i]);
            _selectionCriterion[i] = NULL;
         }
         delete[](_selectionCriterion);
         for (int i = 0; i<_nRoots; ++i) {
            delete(_residuals[_Na-1][i]);
         }
         delete[](_residuals[_Na-1]);
         for (int i = 0; i<_Na-1; ++i) {
            for (int j = 0; j<_nNodes[i]; ++j) {
               _residuals[i][j]->setArrayNull();
               delete(_residuals[i][j]);
               _residuals[i][j] = NULL;
            }
            delete[](_residuals[i]);
         }
         delete[](_residuals);
         delete[](_n);
         delete[](_sizeEdges);
         for (int i = 0; i<_Na; ++i) {
            for (int j = 0; j<_nNodes[i]; ++j) {
               _dicts[i][j] = NULL;
            }
            delete[](_dicts[i]);
            _dicts[i] = NULL;
         }
         delete[](_dicts);

         delete[](_nNodes);
         for (int i = 0; i<_Nt; ++i) {
            delete[](_thrshold[i]);
            _thrshold[i] = NULL;
         }
         delete[](_thrshold);
      };

   /// initialize the processor.
   template <typename T, int V>
      void Processor<T,V>::initialize(int indexInCoeffs,
            int indexInImage) {
         _indexPatch = indexInCoeffs;
         _indexInImage = indexInImage;
         /// initialize the residual
         /// WARNING, between Na and Nt, the quadtree structure does 
         /// not apply anymore.
         int n = static_cast<int>(std::sqrt(static_cast<T>(_nRoots)));
         for (int i = 0; i<n; ++i) {
            for (int j = 0; j<n; ++j) {
               _image->setPatch(*(_residuals[_Na-1][i*n+j]),
                     indexInImage+(i*_image->width()+j)*_sizeEdges[_Na-1],
                     _sizeEdges[_Na-1], _Na);
            }
         }

         if (_gamma > EPSILON) {
            for (int i = 0; i<_Na; ++i) {
               for (int j = 0; j<_nNodes[i]; ++j) {
                  _dicts[i][j] = const_cast<Matrix*>(
                        &(_dictionary->altDicts(i)));
               }
            }
         } else {
            for (int i = 0; i<_Na; ++i) {
               for (int j = 0; j<_nNodes[i]; ++j) {
                  _dicts[i][j] = const_cast<Matrix*>(
                        &(_dictionary->dicts(i)));
               }
            }
         }
         for (int i = 0; i<_Na; ++i) {
            for (int j = 0; j<_nNodes[i]; ++j) {
               _invNorm[i][j] = 
                  const_cast<T**>(_dictionary->invNorm())[i];
            }
         }

         initializeAux();
      };

   /// initialize the processor.
   template <typename T, int V>
      void Processor<T,V>::initialize(int indexInCoeffs,
            const VectR& patch) {
         assert(check(patch));
         _indexPatch = indexInCoeffs;
         /// WARNING, a call to the simple function "copy", 
         /// will break the shared memory structure
         _residuals[_Na-1][0]->copyPart(patch,0);
         for (int i = 0; i<_Na; ++i) {
            for (int j = 0; j<_nNodes[i]; ++j) {
               _dicts[i][j] = const_cast<Matrix*>(
                     &(_dictionary->dicts(i)));
            }
         }
         for (int i = 0; i<_Na; ++i) {
            for (int j = 0; j<_nNodes[i]; ++j) {
               _invNorm[i][j] = 
                  const_cast<T**>(_dictionary->invNorm())[i];
            }
         }
         initializeAux();
      };

   /// Auxiliary function for initialization
   template <typename T, int V>
      void Processor<T,V>::initializeAux() {
         if (_gamma > EPSILON) {
            for (int i = 0; i<_nNodes[0]; ++i) {
               switchPatchToRGBMetric<T,V>(*(_residuals[0][i]),0,_gamma);
            }
         }
         for (int i = 0; i<_nRoots; ++i) {
            _originalPatch[i]->copy(*(_residuals[_Na-1][i]));
         }
         /// clean everything for a new approximation
         for (int j = 0; j<_nRoots; ++j) {
            _numSelectedAtoms[j] = 0;
         }
         for (int i = 0; i<_nRoots; ++i) {
            delete(_Q[i]);
            _Q[i] = NULL;
            delete(_R[i]);
            _R[i] = NULL;
         }
      };

   /// launch the approximation
   template <typename T, int V>
      void Processor<T,V>::run() {

         /// approximation at each root
         for (int i = 0; i<_nRoots; ++i) {
            nodeApproximation(i);
         }

         /// refine approximation
         if (_Nt > _Na) rootApproximation();
         /// Compute the coefficients alpha
         computeCoeffs();
         /// insert coefficients in the dictionary
         updateDictionary();
      };

   /// approximation for a node
   template <typename T, int V>
      void Processor<T,V>::nodeApproximation(const int indexNode) {
         bool& converged = _converged[indexNode];
         int& l = _numSelectedAtoms[indexNode];

         /// WARNING, this is only in the case of denoising
         /// when convergence is reached from the first iteration,
         /// the results are better when one force the 
         /// decomposition over the constant atom
         converged = nrm2(*(_residuals[_Na-1][indexNode]))
            <= _thrshold[_Na-1][indexNode] || l >= _maxL;

         if (converged && l <_maxL) {
            _levelOfSelectedAtoms[indexNode][l] = _Na-1;
            _indexSelectedAtoms[indexNode][l] = _dictionary->K(_Na-1)-1;
            _selectedPatches[indexNode][l] = indexNode;
            computeQRResidual(indexNode);
            return;
         }

         while (!converged) {
            /// compute the scores
            updateCriterion(indexNode, _Na-1);

            /// select the best atom
            int& levelSelected = _levelOfSelectedAtoms[indexNode][l];
            int& atomSelected = _indexSelectedAtoms[indexNode][l];
            int& nodeSelected = _selectedPatches[indexNode][l];
            updateScores(indexNode, _Na-1,levelSelected,nodeSelected,
                  atomSelected);

            /// orthogonalize, update Q, R, residual
            computeQRResidual(indexNode, levelSelected,nodeSelected,
                  atomSelected);

            /// test the convergence
            converged = nrm2(*(_residuals[_Na-1][indexNode]))
               <= _thrshold[_Na-1][indexNode] || l >= _maxL;
         }
      };

   /// approximation for the root, refinement of the previous results
   template <typename T, int V>
      void Processor<T,V>::rootApproximation() {
         T* normsRoot = new T[_nRoots];
         for (int i = 0; i<_nRoots; ++i) {
            int& l = _numSelectedAtoms[i];
            if (l >= _maxL) {
               delete[](normsRoot);
               return;
            }
            updateCriterion(i,_Na-1);
            updateScores(i,_Na-1);
            normsRoot[i] = *(_residuals[_Na-1][i]) %
               *(_residuals[_Na-1][i]);
         }

         for (int i = _Na; i<_Nt; ++i) {
            for (int j = 0; j<_nNodes[i]; ++j) {
               rootApproximation(i,j,normsRoot);
            }
         }
         delete[](normsRoot);
      };

   /// approximation for the root, refinement of the previous results
   template <typename T, int V>
      void Processor<T,V>::rootApproximation(const int level, 
            const int node, T* normsRoot) {
         assert(level >= _Na-1);
         T thrshold = _thrshold[level][node]*_thrshold[level][node];
         int numNodes = fourPow[level-_Na+1];
         int minNode = numNodes*node;
         int maxNode = numNodes*(node+1)-1;
         VectR* norm = new VectR(numNodes);
         for (int i = minNode; i<=maxNode; ++i) {
            (*norm)(i-minNode) = normsRoot[i];
         }
         /// check global convergence
         bool globalConvergence = sum(*norm) <= thrshold;
         if (globalConvergence) {
            delete(norm);
            return;
         }

         VectR* scores = new VectR(numNodes);
         /// compute the scores
         for (int i = 0; i<=maxNode-minNode; ++i) {
            (*scores)(i) = _score[_Na-1][i+minNode];
         }

         while (true) {
            /// - pick the best score
            int chosenRoot = idamax(*scores)+minNode;
            int &l = _numSelectedAtoms[chosenRoot];

            /// perform the orthogonalization
            if (l+1 >= _maxL) break;
            computeQRResidual(chosenRoot);

            /// convergence test
            (*norm)(chosenRoot-minNode) = 
               *(_residuals[_Na-1][chosenRoot]) %
               *(_residuals[_Na-1][chosenRoot]);
            globalConvergence =   sum(*norm) <= thrshold;
            normsRoot[chosenRoot] = (*norm)(chosenRoot-minNode);

            /// update the different variables
            updateCriterion(chosenRoot,_Na-1);
            updateScores(chosenRoot,_Na-1);
            if (globalConvergence) break;
            (*scores)(chosenRoot-minNode) = _score[_Na-1][chosenRoot];
         }

         delete(norm);
         delete(scores);
      };

   /// update the dictionary
   template <typename T, int V>
      void Processor<T,V>::updateDictionary() {
         for (int j = 0; j<_nRoots; ++j) {
            int& l = _numSelectedAtoms[j];
            for (int k = 0; k<l; ++k) {
               int& level = _levelOfSelectedAtoms[j][k];
               _dictionary->insertCoef(level,
                     _indexPatch*_nNodes[level]+_selectedPatches[j][k],
                     _indexSelectedAtoms[j][k],
                     _coeffs[j][k]*
                     _invNorm[level][_selectedPatches[j][k]]
                     [_indexSelectedAtoms[j][k]]);
            }
         }
         int index = _indexPatch*_nRoots;
         if (_dictionary->storeResidual()) {
            if (_gamma > EPSILON) {
               for (int i = 0; i<_nNodes[0]; ++i) {
                  switchPatchBackFromRGBMetric<T,V>(
                        *(_residuals[0][i]),0,_gamma);
               }
            }
            for (int i = 0; i<_nRoots; ++i) {
               _dictionary->updateResidual(index+i,
                     *(_residuals[_Na-1][i]));
            }
         } 
         if (_image != NULL) {
            int n = static_cast<int>(std::sqrt(static_cast<T>(_nRoots)));
            for (int i = 0; i<n; ++i) {
               for (int j = 0; j<n; ++j) {
                  _dictionary->insertIndexPatch(_indexInImage+
                        (i*_image->width()+j)*_sizeEdges[_Na-1],
                        index+i*n+j);
               }
            }
         }
      };


   /// update the scores for one root and its children
   template <typename T, int V>
      inline void Processor<T,V>::updateCriterion(const int indexNode,
            const int level) {
         *(_selectionCriterion[level][indexNode]) =
            *(_residuals[level][indexNode]) * 
            (*(_dicts[level][indexNode]));
#ifdef NNEG
         for (int i = 0; i<_dictionary->K(level); ++i) {
            if ((*(_selectionCriterion[level][indexNode]))(i) < 0)
               (*(_selectionCriterion[level][indexNode]))(i) = 0;
         }
#endif
         if (_weightConstant) {
            for (int i = 0; i<V; ++i) {
               (*(_selectionCriterion[level]
                  [indexNode]))(_dictionary->K(level)-V+i) *= WEIGHT_CONSTANT;
            }
         }
         if (level > 0) {
            for (int i = 0; i<4; ++i) {
               updateCriterion(4*indexNode+i,level-1);
            }
         }
      };


   /// update the scores for each node and select the best
   /// level and atom
   template <typename T, int V>
      inline void Processor<T,V>::updateScores(const int indexNode,
            const int level) {
         int& l = _numSelectedAtoms[indexNode];
         assert(l < _maxL);
         int& levelSelected = _levelOfSelectedAtoms[indexNode][l];
         int& atomSelected = _indexSelectedAtoms[indexNode][l];
         int& nodeSelected = _selectedPatches[indexNode][l];
         updateScores(indexNode,level,levelSelected,nodeSelected,
               atomSelected);
      };


   /// update the scores for each node and select the best
   /// level and atom
   template <typename T, int V>
      inline void Processor<T,V>::updateScores(const int indexNode, 
            const int level, int& levelSelected,int& nodeSelected,
            int& atomSelected) {
         levelSelected = level;
         atomSelected = idamax(*(_selectionCriterion[level][indexNode]));
         nodeSelected = indexNode;
         _score[level][indexNode] = 
            fabs((*(_selectionCriterion[level][indexNode]))
                  (atomSelected));
         if (level > 0) {
            vector<int> levelSelectedChildren(4);
            vector<int> atomSelectedChildren(4);
            vector<int> nodeSelectedChildren(4);
            for (int i = 0; i<4; ++i) {
               updateScores(4*indexNode+i,level-1,
                     levelSelectedChildren[i], nodeSelectedChildren[i],
                     atomSelectedChildren[i]);
               if (_score[level-1][4*indexNode+i] > 
                     _score[level][indexNode]) {
                  levelSelected = levelSelectedChildren[i];
                  atomSelected = atomSelectedChildren[i];
                  nodeSelected = nodeSelectedChildren[i];
                  _score[level][indexNode] = 
                     _score[level-1][4*indexNode+i];
               }
            }
         }
      };

   /// compute the coefficients alpha
   template <typename T, int V>
      inline void Processor<T,V>::computeCoeffs() {
         for (int i = 0; i<_nRoots; ++i) {
            int l = _numSelectedAtoms[i];
            if (l != 0) {
               VectR* tmp = new VectR((*(_originalPatch[i])) *
                     (*(_Q[i])));
               dcopy_(l,tmp->array,1,_coeffs[i],1);
               delete(tmp);
               dtrsv_('u','n','n',l,_R[i]->array,l,_coeffs[i],1);
            }
         }
         //  refinementCoeffs();
      };

   /// update the matrices Q, R and residual
   template <typename T, int V>
      inline void Processor<T,V>::computeQRResidual(
            const int indexNode) {
         int& l = _numSelectedAtoms[indexNode];
         int& levelSelected = _levelOfSelectedAtoms[indexNode][l];
         int& atomSelected = _indexSelectedAtoms[indexNode][l];
         int& nodeSelected = _selectedPatches[indexNode][l];
         computeQRResidual(indexNode,levelSelected,nodeSelected,
               atomSelected);
      };

   /// update the matrices Q, R and residual
   template <typename T, int V>
      inline void Processor<T,V>::computeQRResidual(const int indexNode,
            const int levelSelected, const int nodeSelected,
            const int atomSelected) {
         int& l = _numSelectedAtoms[indexNode];
         VectR* atom;
         atom = new VectR(_dicts[levelSelected][nodeSelected]->
               colRef(atomSelected));
         int shift = (nodeSelected % fourPow[_Na-1-levelSelected])
            *_n[levelSelected];
         if (l == 0) {
            _Q[indexNode] = new Matrix(_n[_Na-1],1);
            _R[indexNode] = new Matrix(1,1);
            VectR w(_n[_Na-1]);
            w.zero();
            w.copyPart(*atom,shift);

            /// The dictionary should be normalized.
            _R[indexNode]->set(0,0,1.0);
            /// do not do a full copy
            _Q[indexNode]->setColumn(w,0);
            double tmp = *(_residuals[_Na-1][indexNode]) % w;
            *(_residuals[_Na-1][indexNode]) -= (tmp * w);
         } else {
            Matrix& Q = *(_Q[indexNode]);
            Matrix& R = *(_R[indexNode]);

            /// prepare for the full scale orthogonalization
            VectR w(Q.m);
            w.zero();
            w.copyPart(*atom,shift);

            VectR r1(l+1);
            for (int i = 0; i<l; ++i) {
               /// do not do a full copy
               VectR qi = _Q[indexNode]->colRef(i);
               r1(i) = qi % w;
               w -= r1(i) * qi;
               qi.setArrayNull();
            }
            T norm2W = nrm2(w);
            R.addOneLineOneColumn(r1);
            /// criterion for reorthogonalization (same as in 
            /// ARPACK), we assume the initial vector normalized
            if (1.0/norm2W >= SQRT2) {
               VectR r2(l+1);
               for (int i = 0; i<l; ++i) {
                  /// do not do a full copy
                  VectR qi = _Q[indexNode]->colRef(i);
                  r2(i) = qi % w;
                  w -= r2(i) * qi;
                  qi.setArrayNull();
               }
               norm2W = nrm2(w);
               R.addToLastColumn(r2);
            }
            R(l,l) = norm2W;
            w /= norm2W;
            Q.addOneColumn(w);
            double tmp = *(_originalPatch[indexNode]) % w;
            *(_residuals[_Na-1][indexNode]) -= (tmp * w);
         }
         atom->setArrayNull();
         delete(atom);
         ++l;
      };

   namespace Weighted {
      template <typename T, int V>
         class Processor : public KSVD::Processor<T,V> {
            public:
               /// Constructor
               Processor(Dictionary<T,V>* dictionary, double minSigma,
                     int N,
                     const int choiceC,
                     const T gamma,
                     const Image<T,V>* const image,
                     const Image<T,V>* const mapSigma,
                     const int maxL = DEFAULT_MAX_SPARSITY,
                     const bool weightConstant = true);

               /// Destructor
               virtual ~Processor();

               /// initialize the processor.
               virtual void initialize(int indexInCoeffs, 
                     int indexInImage);

            private:

               /// the matrix \f$ \beta \f$
               const Image<T,V>* const _mapSigma;
               /// the minimum value in mapSigma
               const T _minSigma;
               /// original thrsholds before modification
               T* _originalThrsholds;
         };

      /// Constructor
      template <typename T, int V>
         Processor<T,V>::Processor(Dictionary<T,V>* dictionary, 
               double minSigma, int N,
               const int choiceC,
               const T gamma,
               const Image<T,V>* const image,
               const Image<T,V>* const mapSigma,
               const int maxL,
               const bool weightConstant) :
            KSVD::Processor<T,V>::Processor(dictionary,
                  minSigma,N,choiceC,gamma,image,
                  maxL,weightConstant),
            _mapSigma(mapSigma),
            _minSigma(minSigma) {
               for (int i = 0; i<this->_Na; ++i) {
                  for (int j = 0; j<this->_nNodes[i]; ++j) {
                     this->_dicts[i][j] = new Matrix(this->_n[i],
                           this->_dictionary->K(i));
                     this->_invNorm[i][j] = new T[this->_dictionary->
                        K(i)];
                  }
               }
               _originalThrsholds = new T[this->_Nt];
               for (int i = 0; i< this->_Nt; ++i)
                  _originalThrsholds[i] = this->_thrshold[i][0];
            };

      /// Destructor
      template <typename T, int V>
         Processor<T,V>::~Processor() {
            for (int i = 0; i<this->_Na; ++i) {
               for (int j = 0; j<this->_nNodes[i]; ++j) {
                  delete(this->_dicts[i][j]);
                  this->_dicts[i][j] = NULL;
                  delete[](this->_invNorm[i][j]);
                  this->_invNorm[i][j] = NULL;
               }
            }
            delete[](_originalThrsholds);
         };

      /// initialize the processor.
      template <typename T, int V>
         void Processor<T,V>::initialize(int indexInCoeffs,
               int indexInImage) {
            this->_indexPatch = indexInCoeffs;
            this->_indexInImage = indexInImage;
            int n = static_cast<int>(std::sqrt(
                     static_cast<T>(this->_nRoots)));
            for (int i = 0; i<n; ++i) {
               for (int j = 0; j<n; ++j) {
                  this->_image->setPatch(*(this->_residuals[this->_Na-1]
                           [i*n+j]),
                        indexInImage+(i*this->_image->width()+j)*
                        this->_sizeEdges[this->_Na-1],
                        this->_sizeEdges[this->_Na-1],
                        this->_Na);
               }
            }
            for (int i = 0; i<this->_Na; ++i) {
               for (int j = 0; j<this->_nNodes[i]; ++j) {
                  this->_dicts[i][j]->copy(this->_dictionary->dicts(i));
               }
            }
            /// zero-norm of the weights
            int** betal0 = new int*[this->_Nt];
            for (int i = 0; i< this->_Nt; ++i) {
               betal0[i] = new int[this->_nNodes[i]];
            }
            for (int i = 0; i<n; ++i) {
               for (int j = 0; j<n; ++j) {
                  VectR weight(this->_n[this->_Na-1]);
                  _mapSigma->setPatch(weight,
                        indexInImage+(i*_mapSigma->width()+j)*
                        this->_sizeEdges[this->_Na-1],
                        this->_sizeEdges[this->_Na-1],
                        this->_Na);
                  betal0[this->_Na-1][i*n+j] = numNonZeros(weight);
                  // element-wise multiplication, modify patch
                  (*(this->_residuals[this->_Na-1][i*n+j])) *= weight;
                  int numNode = i*n+j;
                  for (int k = 0; k<this->_Na; ++k) {
                     int tmp = fourPow[this->_Na-1-k];
                     int newL = weight.l/tmp;
                     for (int l = 0; l<tmp; ++l) {
                        VectR quad(newL,weight,l);
                        if (isConstant(quad)) {
                           if (this->_gamma > EPSILON) {
                              this->_dicts[k][numNode*tmp+l]->copy(
                                    this->_dictionary->altDicts(k));
                           }
                           T invWeight = fabs(weight(0)) < EPSILON ?
                              1.0 : 1.0/weight(0);
                           for (int m = 0; m<
                                 this->_dictionary->K(k); ++m) {
                              this->_invNorm[k][numNode*tmp+l][m] =
                                 invWeight*
                                 ((this->_dictionary->invNorm())[k][m]);
                           }
                        } else {
                           applyWeightToDict<T>(*(this->_dicts[k]
                                    [numNode*tmp+l]),quad);
                           if (this->_gamma > EPSILON) {
                              switchDictToRGBMetric<T,V>(*(this->_dicts[k]
                                       [numNode*tmp+l]),
                                    this->_n[0],fourPow[k],this->_gamma);
                           }
                           normalizeDict(*(this->_dicts[k]
                                    [numNode*tmp+l]),
                                 this->_invNorm[k][numNode*tmp+l]);
                        }
                        quad.setArrayNull();
                     }
                  }
               }
            }
            for (int i = this->_Na; i<this->_Nt; ++i) {
               for (int j = 0; j<this->_nNodes[i]; ++j) {
                  betal0[i][j] = 0;
                  for (int k = 0; k<4; ++k) {
                     betal0[i][j] += betal0[i-1][4*j+k];
                  }                  
               }
            }
            for (int i = this->_Na-1; i<this->_Nt; ++i) {
               for (int j = 0; j<this->_nNodes[i]; ++j) {
                  this->_thrshold[i][j] = 
                     _originalThrsholds[i]*betal0[i][j]/this->_n[i];
               }
            }
            for (int i = 0; i<this->_Nt; ++i) {
               delete[](betal0[i]);
               betal0[i] = NULL;
            }
            delete[](betal0);
            this->initializeAux();
         };
   };
};

#endif // PROCESSOR_H
