/* trainer.h */
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright (C) 2007 Julien Mairal 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.
   */

/*!
 * \file
 *                Multiscale KSVD Denoising
 *
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                      File trainer.h
 * \brief Contains the class Trainer */

#ifndef TRAINER_H
#define TRAINER_H

#include "processor.h"
#include "video.h"
#define FREQUENCY_WRITE_DICTIONARY 10
#ifdef _OPENMP
#include <omp.h>
#endif

using namespace std;

namespace KSVD {

   /*! \brief 
    * namespace reserved for the training of a dictionary
    * */
   namespace Trainer {

      /*!
       * \brief This class implements Trainer for learning dictionaries on a generic
       * database of images.
       **/
      template <typename T = double, int V = 1> class Trainer {
         public:
            /// Constructor, fileName is the intial dictionary to train
            Trainer(const QString& fileName,
                  const QString& outputDictionary,
                  int dataPerImage, int L,int J);

            /// Constructor for external interfaces
            Trainer(const QString& fileNameInitialDictionary,
                  const QString& fileNameOutputDictionary,
                  const QString& fileNameData,
                  int numData, int L, int J, bool writeOutputDictionary = true);

            /// Constructor for external interfaces
            Trainer(Dictionary<T,V>* dict,
                  Matrix* dataBase, int L, int J);

            /// Constructor for external interfaces
            Trainer(Dictionary<T,V>* dict,
                  Matrix* dataBase, double err, int J);

            /// Destructor
            ~Trainer();

            /// add one image to the database
            void inline addToDatabase(const QString& fileName);
            /// launch the training
            void inline run();

            /// reverse the structure of the database
            void reverseStructureDataBase();

            /// function to call when the dictionary and the dataBase are provided
            void inline setDictDataBaseNull() {
               _dict=NULL;
               _dataBase=NULL;
            };

         protected:
            /// auxiliary function for run
            void runAux();

            /// name of the file for the output dictionary
            const QString& _outputDictionary;
            /// The dictionary
            Dictionary<T,V>* _dict;
            /// the database of images (fileName)
            std::list<QString*>* _imageDataBase;
            /// database
            Matrix* _dataBase;
            /// the sparsity of the decomposition
            int _L;
            /// the minimum residual error
            double _err;
            /// num patches per Image
            int _dataPerImage;
            /// number of iterations
            int _J;
            /// number of patches
            int _numPatches;

         private:
            /// build the database and fill _dataBase
            void inline buildDataBase();

            /// should build database or not
            bool _shouldBuildDataBase;
            /// should write the output dictionary or not
            bool _shouldWriteDictionary;
            /// should write the output dictionary as an image
            bool _shouldWriteImageDictionary;
            /// should write coeffs or not
            bool _shouldWriteCoeffs;
            /// should erase the coefficients at the end
            bool _shouldTerminateLearning;
            /// external interface 
            bool _externalInterface;

      };

      /// Constructor
      template<typename T, int V>
         Trainer<T,V>::Trainer(const QString& fileName, 
               const QString& outputDictionary,
               int dataPerImage, int L, int J) :
            _L(L), _J(J), _dataBase(NULL),
            _err(0),
            _outputDictionary(outputDictionary),
            _shouldBuildDataBase(true),
            _dataPerImage(dataPerImage),
            _shouldWriteDictionary(true),
            _shouldWriteCoeffs(false),
            _shouldTerminateLearning(true),
            _externalInterface(false),
            _shouldWriteImageDictionary(true),
            _numPatches(0) {
               _dict = new Dictionary<T,V>(fileName,true); 
               _imageDataBase = new std::list<QString*>();
            };

      /// Constructor for external interfaces
      template<typename T, int V>
         Trainer<T,V>::Trainer(const QString& fileNameInitialDictionary,
               const QString& fileNameOutputDictionary,
               const QString& fileNameData,
               int numData, int L, int J, bool writeOutputDictionary) :
            _L(L), _J(J),
            _err(0.0),
            _outputDictionary(fileNameOutputDictionary),
            _dataPerImage(0),
            _shouldWriteDictionary(writeOutputDictionary && _J > 0),
            _shouldBuildDataBase(false),
            _shouldWriteCoeffs(true),
            _externalInterface(true),
            _shouldWriteImageDictionary(false),
            _numPatches(numData) {
               _dict = new Dictionary<T,V>(fileNameInitialDictionary,false);
               _imageDataBase = new std::list<QString*>();
               int sizeEdge = _dict->sizeEdgeBigPatches();
               int sizePatches = sizeEdge*sizeEdge*V;
               _dataBase = new Matrix(sizePatches,_numPatches);
               std::ifstream s( fileNameData );
               for (int i = 0; i<_numPatches; ++i) {
                  for (int j = 0; j<sizePatches; ++j) {
                     s >> (*_dataBase)(j,i);
                  }
               }
               for (int i = 0; i<_numPatches; ++i) {
                  VectR col = _dataBase->colRef(i);
                  CPPL::changeStructurePatch<T,V>(col,sizeEdge,_dict->N()-1);
                  col.setArrayNull();
               }
            };


      /// Constructor for external interfaces
      template<typename T, int V>
         Trainer<T,V>::Trainer(Dictionary<T,V>* dict,
               Matrix* dataBase,
               int L, int J) :
            _L(L), _J(J), 
            _err(0.0),
            _outputDictionary("tmpDict"),
            _dataPerImage(0),
            _shouldWriteDictionary(false),
            _shouldBuildDataBase(false),
            _externalInterface(true),
            _shouldWriteCoeffs(false),
            _shouldTerminateLearning(false),
            _dataBase(dataBase),
            _dict(dict),
            _shouldWriteImageDictionary(false) {
               _numPatches=_dataBase->n;
#ifdef NRANDOM
               srand(0);
#else
               srand( (unsigned) time(NULL));
#endif
               _imageDataBase = new std::list<QString*>();
               int sizeEdge = _dict->sizeEdgeBigPatches();
               int sizePatches = sizeEdge*sizeEdge*V;
               for (int i = 0; i<_numPatches; ++i) {
                  VectR col = _dataBase->colRef(i);
                  CPPL::changeStructurePatch<T,V>(col,sizeEdge,_dict->N()-1);
                  col.setArrayNull();
               }
            };

      /// Constructor for external interfaces
      template<typename T, int V>
         Trainer<T,V>::Trainer(Dictionary<T,V>* dict,
               Matrix* dataBase,
               double err, int J) :
            _err(err),
            _J(J), 
            _outputDictionary("tmpDict"),
            _dataPerImage(0),
            _shouldWriteDictionary(false),
            _shouldBuildDataBase(false),
            _shouldWriteCoeffs(false),
            _shouldTerminateLearning(false),
            _externalInterface(true),
            _dataBase(dataBase),
            _dict(dict),
            _shouldWriteImageDictionary(false) {
               _numPatches=_dataBase->n;
               _L = min(DEFAULT_MAX_SPARSITY,_dict->K(0));
               _imageDataBase = new std::list<QString*>();
               int sizeEdge = _dict->sizeEdgeBigPatches();
               int sizePatches = sizeEdge*sizeEdge*V;
#ifdef NRANDOM
               srand(0);
#else
               srand( (unsigned) time(NULL));
#endif
               for (int i = 0; i<_numPatches; ++i) {
                  VectR col = _dataBase->colRef(i);
                  CPPL::changeStructurePatch<T,V>(col,sizeEdge,_dict->N()-1);
                  col.setArrayNull();
               }
            };


      /// Destructor
      template<typename T, int V>
         Trainer<T,V>::~Trainer() {
            delete(_dict);
            delete(_dataBase);
            for (std::list<QString*>::iterator it = 
                  _imageDataBase->begin(); it != _imageDataBase->end();
                  ++it) {
               delete(*it);
               *it = NULL;
            }
            delete(_imageDataBase);
         };

      /// add a file to the dataBase
      template <typename T, int V>
         void inline Trainer<T,V>::addToDatabase(
               const QString& fileName) {
            QString* newImage = new QString(fileName);
            _imageDataBase->push_back(newImage);
         };

      /// launch the training algorithm for building a dictionary
      template <typename T, int V>
         void inline Trainer<T,V>::run() {
            if (_shouldBuildDataBase) buildDataBase();
            runAux();
         };


      /// launch the training algorithm for building a dictionary
      template <typename T, int V>
         void inline Trainer<T,V>::runAux() {
#ifndef NDEBUG
            if (!check(*_dataBase)) cerr << "Error database" << endl;
#endif
            _dict->allocateLearning(_numPatches,_dict->N(),true,true);
#ifdef _OPENMP
            int NUM_THREADS =  omp_get_num_procs();
            Processor<T,V>** processors = new Processor<T,V>*[NUM_THREADS];
            for (int i = 0; i<NUM_THREADS; ++i) {
               if (_err == 0.0) {
                  processors[i] = new Processor<T,V>(_dict,EPSILON,
                        _dict->N(),93,0,NULL,_L,V<=1);
               } else {
                  processors[i] = new Processor<T,V>(_dict,_err,
                        _dict->N(),-1,0,NULL,_L,V<=1);
               }
            }
            for (int i = 0; i<=_J; ++i) {
               if (_J > 0) cout << "Iteration " << i << endl; 
               if (i > 0) {
                  cout << "Updating dictionary" << endl;
                  _dict->performLearning(2*i <= _J);
                  _dict->resetLearning();
                  if (_shouldWriteDictionary &&
                        i % FREQUENCY_WRITE_DICTIONARY == 0) {
                     _dict->writeDictionary(_outputDictionary);
                  }
               }
               if (i < _J || _externalInterface || _J==0) {
                  VectR* patch;
                  int jj;
                  int j = 0;
//                  omp_set_num_threads(NUM_THREADS);
                  omp_set_nested(false);
                  omp_set_dynamic(true);
#pragma omp parallel private(patch,jj) shared(j)
                  {
#pragma omp for schedule(static,1)
                     for (int k = 0; k<NUM_THREADS;++k) {
                        while (j<_numPatches) {
#pragma omp critical
                           {
#pragma omp flush(j)
                              jj = j;
                              ++j;
                           } // End of critical
                           if (jj <_numPatches) {
                              patch = new VectR(_dataBase->colRef(jj));
                              processors[k]->initialize(jj,*patch);
                              processors[k]->run();
                              patch->setArrayNull();
                              delete(patch);
                           }
                        }
                     }
                  } // End of parallel
               }
            }
            for (int i = 0; i<NUM_THREADS; ++i) {
               delete(processors[i]);
            }
            delete[](processors);
#else
            Processor<T,V>* processor;
            if (_err == 0.0) {
               processor = new Processor<T,V>(_dict,
                     EPSILON,
                     _dict->N(),93,0,NULL,_L,V<=1);
            } else {
               processor = new Processor<T,V>(_dict,_err,
                     _dict->N(),-1,0,NULL,_L,V<=1);
            }
            for (int i = 0; i<=_J; ++i) {
               if (_J > 0) cout << "Iteration " << i << endl; 
               if (i > 0) {
                  _dict->performLearning(2*i <= _J);
                  _dict->resetLearning();
                  if (_shouldWriteDictionary && i % FREQUENCY_WRITE_DICTIONARY == 0) {
                     _dict->writeDictionary(_outputDictionary);
                  }
               }
               if (i < _J || _externalInterface || _J==0) {
                  for (int j = 0; j<_numPatches; ++j) {
                     VectR* patch = new VectR(_dataBase->colRef(j));
                     assert(check(*patch));
                     processor->initialize(j,*patch);
                     processor->run();
                     patch->setArrayNull();
                     delete(patch);
                  }
               }
            }
            delete(processor);
#endif
            if (_shouldWriteCoeffs) {
               _dict->writeCoeffs(_outputDictionary+".coeff");
            }
            if (_shouldTerminateLearning) {
               _dict->terminateLearning();
            }
            if (_shouldWriteDictionary) {
               _dict->writeDictionary(_outputDictionary);
               if (_shouldWriteImageDictionary) {
                  _dict->writeDictionaryToImage(_outputDictionary+".png");
               }
            }
         };

      /// reverse the structure of the database
      template <typename T, int V>
         void Trainer<T,V>::reverseStructureDataBase() {
            for (int i = 0; i<_numPatches; ++i) {
               VectR tmp = _dataBase->colRef(i);
               CPPL::reverseStructurePatch<T,V>(tmp,_dict->N()-1);
               tmp.setArrayNull();
            }
         };

      /// build the database of images
      template <typename T, int V>
         void inline Trainer<T,V>::buildDataBase() {
            int numImages = _imageDataBase->size();
            _numPatches = numImages * _dataPerImage;
            int sizeEdge = _dict->sizeEdgeBigPatches();
            int sizePatches = sizeEdge*sizeEdge*V;
            _dataBase = new Matrix(sizePatches,_numPatches);
            int count = 0;
            for (std::list<QString*>::iterator it = 
                  _imageDataBase->begin();
                  it != _imageDataBase->end(); ++it) {
               Image<T,V>* currentImage = new Image<T,V>(*(*it));
               int range = (currentImage->width()-sizeEdge+1)*
                  (currentImage->height()-sizeEdge+1);
               for (int j = 0; j<_dataPerImage; ++j) {
                  long num = rand() % range;
                  VectR* patch = new VectR(_dataBase->colRef(count));
                  currentImage->setPatch(*patch,num,sizeEdge,
                        _dict->N());
                  patch->setArrayNull();
                  delete(patch);
                  ++count;
               }
               delete(currentImage);
            }
         };

      namespace Sequence {

         /*!
          * \brief This class implements Trainer for learning dictionaries on a generic database of videos.
          **/
         template <typename T = double, int V = 1, int W = 3>
            class Trainer :
               public KSVD::Trainer::Trainer<T,V*W> {
                  public:
                     /// Constructor
                     Trainer(const QString& fileName,
                           const QString& outputDictionary,
                           int dataPerImage, int L, int J);
                     /// Destructor
                     ~Trainer();

                     /// launch the training
                     void inline run();
                     /// add one video to the database
                     void inline addToDatabase(const QString& baseName,
                           int Nimages);

                  private:
                     /// add one image to the database
                     /// forbid to call this function, which 
                     /// will not be implemented
                     void inline addToDatabase(const QString& fileName);

                     /// build the database and fill _dataBase
                     void inline buildDataBase();

                     /// number of images per video
                     std::list<int>* _numImagesPerVideo;
               };

         /// Constructor
         template<typename T, int V, int W>
            Trainer<T,V,W>::Trainer(const QString& fileName,
                  const QString& outputDictionary,
                  int dataPerImage, int L, int J) : 
               KSVD::Trainer::Trainer<T,V*W>(fileName, outputDictionary,
                     dataPerImage, L, J) {
                  _numImagesPerVideo = new std::list<int>();
               };

         /// Destructor
         template<typename T, int V, int W>
            Trainer<T,V,W>::~Trainer() {
               delete(_numImagesPerVideo);
            };

         /// add a file to the dataBase
         template <typename T, int V, int W>
            void inline Trainer<T,V,W>::addToDatabase(
                  const QString& fileName, const int Nimages) {
               QString* newImage = new QString(fileName);
               this->_imageDataBase->push_back(newImage);
               _numImagesPerVideo->push_back(Nimages);
            };

         /// launch the training algorithm for building a dictionary
         template <typename T, int V, int W>
            void inline Trainer<T,V,W>::run() {
               buildDataBase();
               this->runAux();
            };

         /// build the database of images
         template <typename T, int V, int W>
            void inline Trainer<T,V,W>::buildDataBase() {
#ifdef NRANDOM
               srand(0);
#else
               srand( (unsigned) time(NULL));
#endif
               const int offset = W/2;
               int numVideo = this->_imageDataBase->size();
               this->_numPatches = numVideo * this->_dataPerImage;
               int sizeEdge = this->_dict->sizeEdgeBigPatches();
               int sizePatches = sizeEdge*sizeEdge*V*W;
               this->_dataBase = new Matrix(sizePatches,this->_numPatches);
               int count = 0;
               std::list<int>::iterator itNum = _numImagesPerVideo->begin();
               for (std::list<QString*>::iterator it = 
                     this->_imageDataBase->begin();
                     it != this->_imageDataBase->end(); ++it) {
                  KSVD::Sequence::Video<T,V,W>* currentVideo = 
                     new KSVD::Sequence::Video<T,V,W>(*(*it),*itNum);
                  for (int j = 0; j<this->_dataPerImage; ++j) {
                     long i = (rand() % (*itNum-2*offset))+offset;
                     currentVideo->setCurrentFrame(i);
                     const Image<T,V*W>* const currentImage(
                           currentVideo->currentFrames());
                     int range = (currentImage->width()-sizeEdge+1)*
                        (currentImage->height()-sizeEdge+1);
                     long num = rand() % range;
                     VectR* patch = 
                        new VectR(this->_dataBase->colRef(count));
                     currentImage->setPatch(*patch,num,sizeEdge,
                           this->_dict->N());
                     patch->setArrayNull();
                     delete(patch);
                     ++count;
                     /// WARNING
                     //   if (count > 15000) break;
                  }
                  ++itNum;
                  delete(currentVideo);
               }
            }
      };
   };

   typedef Trainer::Trainer<double,3> TrainerRGB;
   typedef Trainer::Trainer<double,1> TrainerGray;
   typedef Trainer::Sequence::Trainer<double,3,3> TrainerVideoRGB;
   typedef Trainer::Sequence::Trainer<double,1,3> TrainerVideoGray;
   typedef Trainer::Sequence::Trainer<double,3,5> TrainerVideoHQRGB;
   typedef Trainer::Sequence::Trainer<double,1,5> TrainerVideoHQGray;

};

#endif // TRAINER_H
