/* video.h */
/* KSVDRestore for denoising, inpainting, demosaicing and scale-up
   Copyright (C) 2007 Julien Mairal 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.
*/

/*!
 * \file
 *                Multiscale KSVD Denoising
 *
 *                      by Julien Mairal
 *                      julien.mairal@m4x.org
 *
 *                      File video.h
 * \brief Contains the class Video. Warning, it has to be used 
 * with small videos since it will load the complete video into
 * memory */

#ifndef VIDEO_H
#define VIDEO_H

#include "image.h"

namespace KSVD {

   namespace Sequence {

      /*!
       * \brief "video" class used for the K-SVD.
       *
       * The elements are of type T, V is the number of channels,
       * W is the number of frames handled simultaneously */
      template <typename T = double, int V = 1, int W = 3> class Video {
         public:
            /// Empty Constructor
            Video();
            /// Constructor without input data
            Video(const int w, const int h, const int Nimages);
            /// Constructor with folder to load
            Video(const QString& inputFolder, const int Nimages);
            /// Copy constructor
            explicit Video(const Video<T,V,W>& video);
            /// Destructor
            ~Video();

            /// print some information about the video
            void printInformation() const;
            /// add a white gaussian noise to the video
            void addNoiseGaussian(const T sigma);
            /// save the video in image files
            bool save(const QString& baseName) const;
            /// set the current frame
            void setCurrentFrame(int newCurrentFrame);
            /// compute the PSNR
            T PSNR(const Video<T,V,W>& vid) const;
            /// erase the data
            void inline clear();
            /// remove a percentage of the data
            Video<T,V,W>* removePercents(const int percent);
            /// apply an inpainting mask
            void applyMap(const Video<T,V,W>& map);
            /// restore the original values using the map
            void restoreFromMap(const Video<T,V,W>& map,
                  const Video<T,V,W>& input);
            /// put a Bayer pattern in the video
            void bayerPattern();
            /// mozaic the video
            void mozaic();
            /// put an interlaced pattern in the video
            void interlacedPattern();
            /// interlace the video
            void interlace();

            /// Accessor 
            inline int width() const { return _width; };
            /// Accessor 
            inline int height() const { return _height; };
            /// Accessor 
            inline int nFrames() const { return _nFrames; };
            /// Accessor
            inline Image<T,V*W>* currentFrames() const {
               return _currentFrames;
            }
            /// Operator []
            inline T& operator[](const int index) { 
               return _dataVideo[index];
            };
            /// Operator []
            inline T operator[](const int index) const {
               return _dataVideo[index]; 
            };

         private:
            /// operator = is forbidden at the moment
            Video<T,V,W>& operator=(const Video<T,V,W>& vid);

            /// width
            int _width;
            /// height
            int _height;
            /// number of frames
            int _nFrames;
            /// number of data
            int _numData;
            /// number of current frame
            int _numCurrentFrame;
            /// current multiFrame Image
            Image<T,V*W>* _currentFrames;
            /// raw data
            T* _dataVideo;
      };

      /// Empty constructor
      template <typename T, int V, int W>
         Video<T,V,W>::Video() : _width(0), _height(0),
         _nFrames(0), _currentFrames(NULL), _dataVideo(NULL),
         _numCurrentFrame(0),
         _numData(0) { };

      /// Constructor without input data
      template <typename T, int V, int W>
         Video<T,V,W>::Video(const int w, 
               const int h, const int Nimages) : 
            _width(w), _height(h), _nFrames(Nimages),
            _numCurrentFrame(0),
            _numData(w*h*Nimages*V) {
               _currentFrames = new Image<T,V*W>(_width,_height);
               _dataVideo = new T[_numData];
            };

      /// Constructor with folder to load
      template <typename T, int V, int W>
         Video<T,V,W>::Video(const QString& fileBase,
               const int Nimages) : _nFrames(Nimages),
         _numCurrentFrame(0) {
            QString file(fileBase+"0.png");
            QImage imageTest(file);
            _width = imageTest.width();
            _height = imageTest.height();
            _numData = V*_width*_height*_nFrames;
            _dataVideo = new T[_numData];
            _currentFrames = new Image<T,V*W>(_width,_height);
            int count = 0;
            for (int i = 0; i<_nFrames; ++i) {
               file = fileBase+QString::number(i)+".png";
               Image<T,V> im(file);
               for (int j = 0; j< im.numData(); ++j) {
                  _dataVideo[count] = im[j];
                  ++count;
               }
            }
         };

      /// Copy constructor
      template <typename T, int V, int W>
         Video<T,V,W>::Video(const Video<T,V,W>& video) :
            _width(video._width), _height(video._height), 
            _numCurrentFrame(0),
            _nFrames(video._nFrames), _numData(video._numData) {
               _currentFrames = new Image<T,V*W>(*video._currentFrames);
               _dataVideo = new T[_numData];
               memcpy(_dataVideo,video._dataVideo,_numData*sizeof(T));
            };

      /// Destructor
      template <typename T, int V, int W>
         Video<T,V,W>::~Video() {
            delete(_currentFrames);
            delete(_dataVideo);
         };

      /// print some information about the video
      template <typename T, int V, int W>
         void Video<T,V,W>::printInformation() const {
            if (V == 1) {
               cout << "Gray Video" << endl;
            } else if (V == 3) {
               cout << "RGB Video" << endl;
            } else {
               cout << "Video" << endl;
            }
            cout << _width << " x " << _height << " x "
               << _nFrames << endl;
         };

      /// add a white gaussian noise to the video
      template <typename T, int V, int W>
         void Video<T,V,W>::addNoiseGaussian(const T sigma) {
            for (int i = 0; i<_numData; ++i) {
               _dataVideo[i] += sigma*normalDistrib<T>();
            }
         };

      /// save the video in image files
      template <typename T, int V, int W>
         bool Video<T,V,W>::save(const QString& baseName) const {
            int count = 0;
            bool res = true;
            for (int i = 0; i< _nFrames; ++i) {
               Image<T,V> image(_width,_height);
               for (int j = 0; j<image.numData(); ++j) {
                  assert(count >= 0 && count < _numData);
                  image[j] = _dataVideo[count];
                  ++count;
               }
               res = res && 
                  image.saveFile(baseName+QString::number(i)+".png");
            }
            return res;
         };

      /// set the current frame
      template <typename T, int V, int W>
         void Video<T,V,W>::setCurrentFrame(int newCurrentFrame) {
            const int sizeFrame = _width*_height*V;
            const int offset = W/2;
            if (newCurrentFrame < offset || 
                  newCurrentFrame > _nFrames-1-offset) {
               cerr << "ERROR in setCurrentFrame " << endl;
               cerr << "Frame : " << newCurrentFrame << " " << _nFrames << " "
                  << offset << endl;
               exit(1);
            }
            int debut = newCurrentFrame-offset;
            for (int i = 0; i<_width*_height; ++i) {
               for (int j = 0; j<W; ++j) {
                  for (int k = 0; k<V; ++k) {
                     (*_currentFrames)[W*V*i+j*V+k] =
                        _dataVideo[(debut+j)*sizeFrame+i*V+k];
                  }
               }
            }
         };

      /// compute the PSNR
      template <typename T, int V, int W>
         T Video<T,V,W>::PSNR(const Video<T,V,W>& input) const {
            T eqm = 0;
            T tmp;
            for (int i = 0; i< input._numData; ++i) {
               tmp = _dataVideo[i]-input._dataVideo[i];
               eqm += tmp*tmp;
            }
            eqm /= static_cast<T>(input._numData);
            const T range = MAX_VALUE-MIN_VALUE;
            return 10.0*log(range*range/eqm)/log(10.0);
         };

      /// erase the data
      template <typename T, int V, int W>
         void Video<T,V,W>::clear() {
            for (int i = 0; i< _numData; ++i) {
               _dataVideo[i] = T();
            }
         };

      /// remove a percentage of the data
      template <typename T, int V, int W>
         Video<T,V,W>* Video<T,V,W>::removePercents(const int percent) {
            int numRemoved = 0;
            int maxRemoved = (_numData*percent)/100;
            Video<T,V,W>* map = new Video<T,V,W>(_width,_height,_nFrames);
            for (int i = 0; i<_numData; ++i) (*map)[i] = 1.0;
            while (numRemoved < maxRemoved) {
               int index = floor(_numData*ran1())+1;
               if (index < _numData && (*map)[index] == 1.0) {
                  numRemoved++;
                  (*map)[index] = 0.0;
                  _dataVideo[index] = -MEAN_VALUE;
               }
            }
            return map;
         };

      /// apply an inpainting mask
      template <typename T, int V, int W>
         void Video<T,V,W>::applyMap(const Video<T,V,W>& map) {
            for (int i = 0; i<_numData; ++i) {
               if (map[i] == 0.0) {
                  _dataVideo[i] = MIN_VALUE-MEAN_VALUE;
               }
            }
         };

      /// apply an inpainting mask
      template <typename T, int V, int W>
         void Video<T,V,W>::restoreFromMap(const Video<T,V,W>& map,
               const Video<T,V,W>& input) {
            for (int i = 0; i<_numData; ++i) {
               if (map[i] == 1.0) {
                  _dataVideo[i] = input[i];
               }
            }
         };

      /// Bayer Pattern
      template <typename T, int V, int W>
         void Video<T,V,W>::bayerPattern() {
            assert(V == 3);
            int count = 0;
            for (int k = 0; k< _nFrames; ++k) {
               for (int i = 0; i< _height; ++i) {
                  for (int j = 0; j< _width; ++j) {
                     bool eq1 = (i % 2 == 0);
                     bool eq2 = (j % 2 == 0);
                     if (eq1 && !eq2) {
                        _dataVideo[count++] = 1.0;
                        _dataVideo[count++] = 0.0;
                        _dataVideo[count++] = 0.0;
                     } else if (eq2 && !eq1) {
                        _dataVideo[count++] = 0.0;
                        _dataVideo[count++] = 0.0;
                        _dataVideo[count++] = 1.0;
                     } else {
                        _dataVideo[count++] = 0.0;
                        _dataVideo[count++] = 1.0;
                        _dataVideo[count++] = 0.0;
                     }
                  }
               }
            }
         };

      /// mozaic the video
      template <typename T, int V, int W>
         void Video<T,V,W>::mozaic() {
            assert(V == 3);
            Video<T,V,W> vid(_width,_height,_nFrames);
            vid.bayerPattern();
            applyMap(vid);
         };

      /// put an interlaced pattern in the video
      template <typename T, int V, int W>
         void Video<T,V,W>::interlacedPattern() {
            int count = 0;
            int even = 0;
            int even2 = 0;
            for (int k = 0; k< _nFrames; ++k) {
               even = 1-even;
               even2 = even;
               for (int i = 0; i< _height; ++i) {
                  even2 = !even2;
                  for (int j = 0; j< _width; ++j) {
                     for (int l = 0; l<V; ++l) {
                        _dataVideo[count++] = even2 ? 1.0 : 0.0;
                     };
                  }
               }
            }
         };

      /// interlace the video
      template <typename T, int V, int W>
         void Video<T,V,W>::interlace() {
            Video<T,V,W> vid(_width,_height,_nFrames);
            vid.interlacedPattern();
            applyMap(vid);
         };

   };

   typedef KSVD::Sequence::Video<double,1,3> VideoGray;
   typedef KSVD::Sequence::Video<double,3,3> VideoRGB;
   typedef KSVD::Sequence::Video<double,1,5> VideoHQGray;
   typedef KSVD::Sequence::Video<double,3,5> VideoHQRGB;
};

#endif // VIDEO_H


